try:
    import importlib.resources as res
except ImportError:  # pragma: no cover
    import importlib_resources as res  # to support python < 3.7


class TestFileMeta(type):
    '''
    This metaclass generates test methods for every file matching given
    rules (as class properties).

    The added test methods will call an existing meta_test method passing
    module name and filename as arguments.

    Functions from :module:`importlib.resources` (see resource_methods) are
    also injected for convenience.

    Honored class properties:
    - meta_test: called with module and filename by injected test methods.
    - meta_module: module to inspect for files
    - meta_file_extensions: filename extensions will result on test injection.
    - meta_prefix: prefix added to injected tests (defaults to `file`)
    '''

    resource_methods = (
        'contents',
        'is_resource',
        'open_binary',
        'open_text',
        'path',
        'read_binary',
        'read_text',
        )

    @classmethod
    def iter_contents(cls, module, extensions):
        for item in res.contents(module):
            if res.is_resource(module, item):
                if any(map(item.endswith, extensions)):
                    yield (module, item)
                continue
            submodule = '%s.%s' % (module, item)
            for subitem in cls.iter_contents(submodule, extensions):
                yield subitem

    @classmethod
    def create_method(cls, module, filename, prefix, extensions):
        def test(self):
            self.meta_test(module, filename)
        strip = max(len(ext) for ext in extensions if filename.endswith(ext))
        test.__name__ = 'test_%s_%s_%s' % (
            prefix,
            module.replace('.', '_').strip('_'),
            filename[:-strip].strip('_'),
            )
        return test

    def __init__(self, name, bases, dct):
        super(TestFileMeta, self).__init__(name, bases, dct)

        # generate tests from files
        name = self.meta_module
        prefix = getattr(self, 'meta_prefix', 'file')
        extensions = self.meta_file_extensions
        for module, item in self.iter_contents(name, extensions):
            test = self.create_method(module, item, prefix, extensions)
            setattr(self, test.__name__, test)

        # add resource methods
        for method in self.resource_methods:
            if not hasattr(self, method):
                setattr(self, method, staticmethod(getattr(res, method)))



# -*- coding: utf-8 -*-

import sys

try:
    import configparser
except ImportError:  # pragma: no-cover
    import ConfigParser as configparser

try:
    import collections as userdict
except ImportError:  # pragma: no-cover
    import UserDict as userdict


UserDict = userdict.UserDict
PY2 = sys.version_info[0] == 2

if PY2:
    byteview = memoryview.tobytes
    StringLike = basestring if PY2 else str  # noqa: F821
    BytesLike = str
    ConfigParser = configparser.SafeConfigParser  # noqa: F821
else:
    def byteview(m):
        return m

    StringLike = str
    BytesLike = (bytes, memoryview)
    ConfigParser = configparser.ConfigParser



# -*- coding: utf-8 -*-

import unittest

import flask

import cookieman
import cookieman.limits
import cookieman.exceptions
import cookieman.cookie


class TestCookie(unittest.TestCase):
    module = cookieman.cookie
    cookie_class = module.Cookie

    def test_partition(self):
        name = 'cookie'
        results = {
            b'asdfg': [
                self.cookie_class(name, b'asdf'),
                self.cookie_class('%s-1' % name, b'g'),
                ],
            b'asdfgh': [
                self.cookie_class(name, b'asdf'),
                self.cookie_class('%s-1' % name, b'gh'),
                self.cookie_class('%s-2' % name),
                ],
            b'asdfgh j': [
                self.cookie_class(name, b'asdfgh'),
                self.cookie_class('%s-1' % name, b' j'),
                self.cookie_class('%s-2' % name),
                ],
            b'asdfgh jk': [
                self.cookie_class(name, b'asdfgh'),
                self.cookie_class('%s-1' % name, b' j'),
                self.cookie_class('%s-2' % name, b'k'),
                ],
            }

        for value, result in results.items():
            cookie = self.cookie_class(name, value)
            maxsize = max(len(r) for r in result) if result else 0
            partitions = list(cookie.partition(maxsize))
            self.assertListEqual(partitions, result)

        with self.assertRaises(cookieman.exceptions.CookieSizeException):
            cookie = self.cookie_class('session', b'asdf')
            list(cookie.partition(5))

        with self.assertRaises(cookieman.exceptions.CookieSizeException):
            cookie = self.cookie_class('session', b' ')
            list(cookie.partition(len(cookie) - 2))

    def test_repr(self):
        cookie = self.cookie_class('name', b'value', {'path': '/res'})
        self.assertEqual(
            repr(cookie),
            "<Cookie 'Set-Cookie: name=value; Path=/res'>"
            )


class TestCookieProcessor(unittest.TestCase):
    class FakeRequest(object):
        def __init__(self, **kwargs):
            for key, value in kwargs.items():
                setattr(self, key, value)

    module = cookieman.cookie
    cookie_class = module.Cookie
    processor_class = module.CookieProcessor
    request_class = FakeRequest

    def request(self, **kwargs):
        return self.request_class(**kwargs)

    def setUp(self):
        self.app = flask.Flask('myapp')
        self.app.session_cookie_name = 'session'
        self.app.session_interface = cookieman.CookieMan()
        self.browser = cookieman.limits.Browser()
        self.processor = self.processor_class(self.app.session_interface)

    def test_partial_cookies(self):
        request = self.request(
            cookies={
                'session': 'asdf',
                'session-1': 'gh',
                'session-2': 'j',
                }
            )
        cookies = list(
            self.processor.iter_request_cookies(
                self.app, request, self.browser))
        self.assertListEqual(cookies, [
            self.cookie_class('session', b'asdf'),
            self.cookie_class('session-1', b'gh'),
            self.cookie_class('session-2', b'j'),
            ])

    def test_bad_partial_cookie_size(self):
        request = self.request(
            cookies={
                'session': 'asdf',
                'session-1': 'asdfasdfasdf',
                },
            )
        cookies = list(
            self.processor.iter_request_cookies(
                self.app, request, self.browser))
        self.assertListEqual(cookies, [self.cookie_class('session', b'asdf')])



# -*- coding: utf-8 -*-

try:
    import importlib.resources as res
except ImportError:  # pragma: no-cover
    import importlib_resources as res


def read_text(resource_name):
    '''
    Get resource

    :param resource_name: name of resource to load
    :type resource_name: str
    :returns: decoded resource content
    :type: str
    '''
    return res.read_text(__name__, resource_name)



# -*- coding: utf-8 -*-

import flask.sessions
import werkzeug.datastructures

import cookieman.compat


class LazySession(cookieman.compat.UserDict, flask.sessions.SessionMixin):
    callback_dict_class = werkzeug.datastructures.CallbackDict

    modified = False
    accessed = False

    def _notify_update(self, data):
        '''
        Mark this object as modified

        :param data: unused
        :type data: Any
        '''
        self.modified = True
        self.accessed = True

    def __init__(self, session_data_fnc, request=None):
        '''
        :param session_data_fnc: session data getter function
        :type session_data_fnc: Callable[[], Mapping[Any, Any]]
        :param request: request object
        :type request: Optional[flask.Request]
        '''
        self._data = None
        self._session_data_fnc = session_data_fnc
        self._request = request

    @property
    def data(self):
        '''
        :returns: returns inner session mapping object
        :rtype: Mapping[Any, Any]
        '''
        self.accessed = True
        if self._data is None:
            self._data = self.callback_dict_class(
                self._session_data_fnc(),
                self._notify_update
                )
        return self._data

    # TODO (on python2 drop): remove
    def __iter__(self):
        '''
        x.__iter__() <==> iter(x)

        :returns: keys in session mapping
        :rtype: Iterator[Any]
        '''
        return iter(self.data)



# -*- coding: utf-8 -*-

__version__ = '0.0.1'

import datetime

import flask.sessions
import flask.helpers

import cookieman.cookie
import cookieman.serializer
import cookieman.session
import cookieman.limits


class CookieMan(flask.sessions.SessionInterface):
    '''
    Compressed-paginated and signed session cookie manager.
    '''

    limits_class = cookieman.limits.Limits
    shrink_class = cookieman.limits.ShrinkManager
    cookie_processor_class = cookieman.cookie.CookieProcessor
    serializer_class = cookieman.serializer.CookieManSerializer
    session_class = cookieman.session.LazySession

    def __init__(self, salt='session-cookie'):
        self._cookie_processor = self.cookie_processor_class(self)
        self._cookie_shrink = self.shrink_class()
        self._limits_cache = {}
        self.salt = salt

    def register(self, key_or_keys, shrink_fnc=None):
        '''
        Register session shrink function for specific session key or
        keys. Can be used as decorator.

        Usage:
        >>> @app.session_interface.register('my_session_key')
        ... def my_shrink_fnc(data):
        ...     del data['my_session_key']
        ...     return data

        :param key_or_keys: key or iterable of keys would be affected
        :type key_or_keys: Union[str, Iterable[str]]
        :param shrink_fnc: shrinking function (optional for decorator)
        :type shrink_fnc: cookieman.abc.ShrinkFunction
        :returns: either original given shrink_fnc or decorator
        :rtype: cookieman.abc.ShrinkFunction
        '''
        return self._cookie_shrink.register(key_or_keys, shrink_fnc)

    def get_browser(self, app, request):
        '''
        Get browser limit object based on app and request.

        :param app: flask app object
        :type app: flask.Flask
        :returns: limits object
        :rtype: cookieman.limits.Browser
        '''
        limits_file = app.config.get('COOKIEMAN_LIMITS_PATH', None)
        if limits_file not in self._limits_cache:
            self._limits_cache[limits_file] = \
                self.limits_class.load(limits_file)
        return self._limits_cache[limits_file].get(
            request.user_agent.browser,
            request.user_agent.platform,
            request.user_agent.version,
            )

    def get_signing_serializer(self, app):
        '''
        Get signing serializer class instance for given app.

        :param app: application instance
        :type app: flask.Flask
        :returns: instance if app has secret key else None
        :rtype: None or serializer class
        '''
        if not app.secret_key:
            return None
        return self.serializer_class(app.secret_key, self.salt)

    def get_expiration_time(self, app, session=None):
        '''
        Get expiration time for a new cookie based on app config

        :param app: flask app
        :type app: flask.Flask
        :param session: unused
        ;returns: expiration time
        :rtype:
        '''
        return datetime.datetime.utcnow() + app.permanent_session_lifetime

    def open_session(self, app, request):
        '''
        Create a lazy session from request cookies.

        :param app: flask app object
        :type app: flask.Flask
        :param request: request object to extract
        :type request: werkzeug.Request
        :returns: session dict-like object
        :rtype: CookieManSession
        '''
        def extract_cookie_data():
            try:
                data = b''.join(
                    cookie.value
                    for cookie in self._cookie_processor.iter_request_cookies(
                        app,
                        request,
                        self.get_browser(app, request),
                        )
                    )
                if not data:
                    return {}
                max_age = flask.helpers.total_seconds(
                    app.permanent_session_lifetime
                    )
                return serializer.loads(data, max_age=max_age)
            except self.serializer_class.bad_sicnature_exception:
                return {}

        serializer = self.get_signing_serializer(app)
        if serializer:
            return self.session_class(extract_cookie_data, request)
        return None

    def save_session(self, app, session, response):
        '''
        Update response with session cookies.

        :param app: flask app object
        :type app: flask.Flask
        :param session: cookieman session object
        :type session: cookieman.session.LazySession
        :param response: flask response object
        :type response: werkzeug.Response
        :returns: given response object (with cookie changes)
        :rtype: werkzeug.Response
        '''
        request = getattr(session, '_request', None) or flask.request
        browser = self.get_browser(app, request)
        cookies = self._cookie_processor

        if session.modified and not session:
            options = cookies.get_cookie_options(app, browser)
            response.delete_cookie(
                app.session_cookie_name,
                domain=options['domain'],
                path=options['path'],
                )
            return

        if session.accessed and session:
            response.vary.add('Cookie')

        if not self.should_set_cookie(app, session):
            return

        iter_cookies = cookies.iter_response_cookies
        cookies = self._cookie_shrink.shrink(
            dict(session),
            browser,
            self.get_signing_serializer(app).dumps,
            lambda payload: iter_cookies(app, payload, browser)
            )

        for cookie in cookies:
            response.set_cookie(cookie.name, cookie.value, **cookie.options)



# -*- coding: utf-8 -*-

import unittest
import json
import tempfile

import werkzeug.useragents

import cookieman.cookie
import cookieman.limits
import cookieman.exceptions


class TestLimits(unittest.TestCase):
    module = cookieman.limits

    def test_database(self):
        data = '\n'.join((
            '[::]',
            'maxcookies = 0',
            'maxsize = 1',
            'maxtotal = 1',
            '',
            '[whatever::]',
            'maxcookies = 0',
            'maxsize = 1',
            'maxtotal = 2',
            '',
            '[whatever:randomOS:]',
            'maxcookies = 0',
            'maxsize = 1',
            'maxtotal = 3',
            '',
            '[whatever:randomOS:3]',
            'maxcookies = 0',
            'maxsize = 1',
            'maxtotal = 4',
            ))
        with tempfile.NamedTemporaryFile() as f:
            f.write(data.encode('utf-8'))
            f.flush()

            db = self.module.Limits.load(f.name)

            browser = db.get('somebrowser', 'os', None)
            self.assertEqual(browser.maxtotal, 1)

            browser = db.get('whatever')
            self.assertEqual(browser.maxtotal, 2)

            browser = db.get('whatever', 'randomOS')
            self.assertEqual(browser.maxtotal, 3)

            browser = db.get('whatever', 'randomOS', (2,))
            self.assertEqual(browser.maxtotal, 3)

            browser = db.get('whatever', 'randomOS', (3,))
            self.assertEqual(browser.maxtotal, 4)

            browser = db.get('whatever', 'randomOS', (4,))
            self.assertEqual(browser.maxtotal, 4)

        self.assertEqual(self.module.Limits(), self.module.Limits())
        self.assertNotEqual(db, self.module.Limits())

    def test_browser(self):
        db = self.module.Limits.load()
        ua = werkzeug.useragents.UserAgent(
            'Mozilla/5.0 (X11; Linux x86_64; rv:65.0) '
            'Gecko/20100101 Firefox/65.0'
            )
        browser = db.get(ua.browser, ua.platform, ua.version)
        self.assertEqual(browser.name, 'firefox')
        self.assertEqual(browser.maxcookies, 150)
        self.assertEqual(browser.maxsize, 4097)
        self.assertEqual(browser.maxtotal, 150 * 4097)

    def test_defaults(self):
        db = self.module.Limits()
        default = db.get()
        self.assertEqual(default.name, '')

        db = self.module.Limits.load()
        firefox = db.get('firefox')
        self.assertEqual(firefox.maxcookies, 50)


class TestBrowser(unittest.TestCase):
    module = cookieman.limits

    def test_repr(self):
        browser = self.module.Browser()
        self.assertIsInstance(repr(browser), str)


class TestShrinker(unittest.TestCase):
    module = cookieman.limits

    def serialize(self, payload):
        return [cookieman.cookie.Cookie('session', payload)]

    def dumps(self, data):
        return json.dumps(data).encode('utf-8')

    def get_cookies(self, data):
        return self.serialize(self.dumps(data))

    def total_len(self, cookies):
        return sum(map(len, cookies))

    def test_register(self):
        manager = self.module.ShrinkManager()
        result = self.get_cookies({})
        maxsize = self.total_len(result)
        browser = self.module.Browser(maxcookies=1, maxsize=maxsize)

        @manager.register(['a', 'b'])
        @manager.register('c')
        def shrink(data, last):
            return {}

        serialized = manager.shrink(
            {'a': 1, 'b': 1}, browser, self.dumps, self.serialize)
        self.assertListEqual(serialized, result)

        serialized = manager.shrink(
            {'a': 1}, browser, self.dumps, self.serialize)
        self.assertListEqual(serialized, result)

        serialized = manager.shrink(
            {'b': 1}, browser, self.dumps, self.serialize)
        self.assertListEqual(serialized, result)

        serialized = manager.shrink(
            {'c': 1}, browser, self.dumps, self.serialize)
        self.assertListEqual(serialized, result)

    def test_shrink(self):
        manager = self.module.ShrinkManager()

        @manager.register('a')
        def shrink(data, last):
            if last or len(data['a']) > 2:
                data['a'] = data['a'][:-1]
            return data

        data = {'a': 'abcdef', 'b': True}
        result = self.get_cookies({'a': 'a'})
        maxsize = self.total_len(result)
        print(maxsize)
        browser = self.module.Browser(maxsize=maxsize)
        serialized = manager.shrink(data, browser, self.dumps, self.serialize)
        self.assertListEqual(serialized, result)

    def test_shrink_priority(self):
        class ShrinkManager(self.module.ShrinkManager):
            def _purge(self, key, data):
                calls.append(key)
                return super(ShrinkManager, self)._purge(key, data)

        manager = ShrinkManager()

        @manager.register('a')
        def sa(data, last):
            calls.append('a')
            if last:
                lastcalls.append('a')
            return data

        @manager.register('b')
        def sb(data, last):
            calls.append('b')
            if last:
                lastcalls.append('b')
            del data['b']
            return data

        @manager.register('c')
        def sc(data, last):
            calls.append('c')
            if last:
                lastcalls.append('c')
            return data

        calls = []
        lastcalls = []
        data = {'a': 'abcdef', 'b': True, 'k': 1, 'c': 2}
        result = self.get_cookies({})
        maxsize = self.total_len(result)
        browser = self.module.Browser(maxsize=maxsize)
        serialized = manager.shrink(data, browser, self.dumps, self.serialize)
        self.assertListEqual(serialized, result)
        self.assertEqual(''.join(calls), 'abckacac')
        self.assertEqual(''.join(lastcalls), 'ac')

    def test_multiple_handlers(self):
        manager = self.module.ShrinkManager()

        @manager.register('a')
        def a1(data, last):
            calls.append(1)
            del data['a']
            return data

        @manager.register('a')
        def a2(data, last):
            calls.append(2)
            del data['a']
            return data

        calls = []
        data = {'a': 1, 'b': 2}
        result = self.get_cookies({})
        maxsize = self.total_len(result)
        browser = self.module.Browser(maxsize=maxsize)
        manager.shrink(data, browser, self.dumps, self.serialize)

        self.assertEqual(len(calls), 1)

    def test_side_effects(self):
        manager = self.module.ShrinkManager()

        @manager.register('a')
        def a(data, last):
            return data

        @manager.register('b')
        def b(data, last):
            if 'a' in data:
                del data['a']  # handled
            if 'c' in data:
                del data['c']  # unhandled
            return data

        data = {'a': 1, 'b': 2, 'c': 3}
        result = self.get_cookies({})
        maxsize = self.total_len(result)
        browser = self.module.Browser(maxsize=maxsize)
        data = manager.shrink(data, browser, self.dumps, self.serialize)

        self.assertEqual(data, result)

    def test_unshrinkable(self):
        manager = self.module.ShrinkManager()

        with self.assertRaises(cookieman.exceptions.CookieSizeException):
            browser = self.module.Browser(maxsize=1)
            manager.shrink({}, browser, self.dumps, self.serialize)

        with self.assertRaises(cookieman.exceptions.CookieSizeException):
            browser = self.module.Browser(maxcookies=0)
            manager.shrink({}, browser, self.dumps, self.serialize)



# -*- coding: utf-8 -*-

import zlib
import base64

import msgpack
import itsdangerous


class CookieManSerializer(object):
    '''
    Signed compressed session cookie serializer.

    This class implements the following:
    - Messagepack python object serialization.
    - Signature check using itsdangerous.TimestampSigner.
    - Compression using python's standard zlib.
    - URL-safe base64 encoding (to http header issues).
    '''

    signer_class = itsdangerous.TimestampSigner
    bad_sicnature_exception = itsdangerous.BadSignature

    def __init__(self, secret, salt):
        '''
        :param secret:
        :type secret:
        :param salt:
        :type salt:
        '''
        self.signer = self.signer_class(secret, salt=salt)

    def dumps(self, data):
        '''
        Get serialized representation of given object.

        :param data: serializable object
        :type data: object
        :returns: serialized data
        :rtype: bytes
        '''
        dumped = msgpack.packb(data, use_bin_type=True)
        signed = self.signer.sign(dumped)
        compressed = zlib.compress(signed)
        return base64.urlsafe_b64encode(compressed)

    def loads(self, data, max_age=None):
        '''
        Get object from serialized data.

        :param data: serialized data
        :type data: bytes
        :returns: deserialized object
        :rtype: object
        '''
        decoded = base64.urlsafe_b64decode(data)
        decompressed = zlib.decompress(decoded)
        signed = self.signer.unsign(decompressed, max_age=max_age)
        return msgpack.unpackb(signed, raw=False)



# -*- coding: utf-8 -*-

import sys
import itertools
import functools
import collections

import werkzeug.datastructures

import cookieman.resources
import cookieman.exceptions
import cookieman.compat


class Browser(object):
    '''
    :attr name:
    :type name: str
    :attr platform:
    :type platform: str
    :attr version:
    :type version: Tuple[Union[int, str]]
    :attr maxcookies:
    :type maxcookies: int
    :attr maxsize:
    :type maxsize: int
    :attr maxtotal:
    :type maxtotal: int
    '''
    __slots__ = (
        'name', 'platform', 'version', 'maxcookies', 'maxsize',
        'maxtotal',
        )

    def __init__(self, name='', platform='', version=(), maxcookies=1,
                 maxsize=4096, maxtotal=None):
        '''
        :param name: browser name
        :type name: str
        :param platform: browser platform
        :type platform: str
        :param version: browser version
        :type version: Iterable[str]
        :param maxcookies: maximum number of cookies
        :type maxcookies: int
        :param maxsize: maximum number of bytes of every cookies
        :type maxsize: int
        :param maxtotal: maximum number of total bytes of all cookies
        :type maxtotal: Union[int, None]
        '''
        self.name = name
        self.platform = platform
        self.version = tuple(version)
        self.maxcookies = maxcookies
        self.maxsize = maxsize
        self.maxtotal = (
            min(maxtotal, maxcookies * maxsize)
            if maxcookies and maxtotal else
            maxtotal
            if maxtotal else
            maxcookies * maxsize
            )

    def __repr__(self):
        '''
        Return repr(self).

        :returns: string representation of object
        :rtype: str
        '''
        name = '%s.%s' % (
            self.__class__.__module__,
            self.__class__.__name__
            )
        props = {attr: getattr(self, attr) for attr in self.__slots__}
        return '<%s%r>' % (name, props)

    @classmethod
    def parse_version(cls, version):
        '''
        Convert version (str or tuple) to version tuple.

        :param version: version value (str or tuple)
        :type version: Union[str, Iterable[str]]
        :returns: version tuple
        :rtype: Tuple[Union[int, str], ...]
        '''
        stringlike = cookieman.compat.StringLike
        if version:
            if isinstance(version, stringlike):
                version = version.split('.')
            return tuple(
                int(i) if isinstance(i, stringlike) and i.isdigit() else i
                for i in version
                )
        return ()

    @classmethod
    def from_spec(cls, name, properties):
        '''
        Create browser object based on limits database data.

        :param name: browser identifier (name:platform:version)
        :type name: str
        :param properties:
        :type properties: Mapping[str, str]
        :returns: browser object
        :rtype: cookieman.limits.Browser
        '''
        name, platform, version = name.split(':')[:3]
        return cls(
            name,
            platform or '',
            cls.parse_version(version),
            int(properties['maxcookies']),
            int(properties['maxsize']),
            int(properties['maxtotal']),
            )


class Limits(object):
    '''
    Browser limits database
    '''
    browser_class = Browser
    mapping_class = werkzeug.datastructures.ImmutableDict
    dbparser_class = cookieman.compat.ConfigParser

    def __init__(self, browsers=()):
        '''
        :param browsers: iterable with all available browsers
        :type browsers: Iterable[cookieman.limits.Browser]
        '''
        self.browsers_by_key = self.mapping_class(
            (key, tuple(browsers)) for key, browsers in itertools.groupby(
                sorted(
                    browsers,
                    key=lambda browser: (browser.name, browser.platform,
                                         browser.version),
                    reverse=True,
                ),
                key=lambda browser: (browser.name, browser.platform)))
        self.default = self.browser_class()

    def get(self, name='', platform='', version=()):
        '''
        Get browser limits object based on given browser name, platform,
        and version.

        :param name: browser name
        :type name: str
        :param platform: browser platform name
        :type platform: str
        :param version: browser version tuple
        :type version: Union[str, tuple[str, ...]]
        :returns: browser limits object
        :rtype: cookieman.limits.Browser
        '''
        version = self.browser_class.parse_version(version)
        search_keys = (
            ((name, platform), version),
            ((name, platform), ()),
            ((name, ''), version),
            ((name, ''), ()),
            (('', ''), ()),  # database default
            )
        for (key, version) in search_keys:
            for browser in self.browsers_by_key.get(key, ()):
                if browser.version <= version:
                    return browser
        return self.default

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__) and
            self.browsers_by_key == other.browsers_by_key
            )

    @classmethod
    def load(cls, path=None):
        '''
        Load browser limits database file. If None or no path is given,
        it will load cookieman's builtin limits database.

        :param path: path or None
        :type path: Union[str, None]
        :returns: Limits object
        :rtype: cookieman.limits.Limits
        '''
        if path is None:
            data = cookieman.resources.read_text('limits.ini')
        else:
            with open(path, 'rb') as f:
                data = f.read().decode('utf-8')

        parser = cls.dbparser_class()
        parser.read_string(data)

        return cls(
            cls.browser_class.from_spec(name, parser[name])
            for name in parser.sections()
            )


class ShrinkManager(object):
    '''
    Manager class for cookie shrinking functions
    '''
    _key_added_exc = type('KeyAdded', (Exception, ), {})

    def __init__(self):
        self.handlers = collections.defaultdict(set)

    def register(self, key_or_keys, shrink_fnc=None):
        '''
        Register session shrink function for specific session key or
        keys. Can be used as decorator.

        Usage:
        >>> @shrinker.register('my_session_key')
        ... def my_shrink_fnc(data):
        ...     del data['my_session_key']
        ...     return data

        :type shrink_fnc: Callable[[dict, bool], dict]

        :param key_or_keys: key or list of keys would be affected
        :type key_or_keys: Union[str, Iterable[str]]
        :param shrink_fnc: shrinking function (optional for decorator)
        :type shrink_fnc: shrink_fnc
        :returns: either original given shrink_fnc or decorator
        :rtype: Union[Callable[[shrink_fnc], shrink_fnc], shrink_fnc]
        '''
        if shrink_fnc is None:
            return functools.partial(self.register, key_or_keys)

        keys = (
            (key_or_keys,)
            if isinstance(key_or_keys, cookieman.compat.StringLike) else
            key_or_keys
            )
        for name in keys:
            self.handlers[name].add(shrink_fnc)
        return shrink_fnc

    @classmethod
    def _purge(cls, key, data):
        '''
        Default handler for unhandled keys. This just removes the key.

        :param key: data dict key to remove
        :type key: str
        :param data: data mapping
        :type data: collections.abc.MutableMapping
        :returns: updated mapping
        :rtype: collections.abc.MutableMapping
        '''
        del data[key]
        return data

    def _iter_handled(self, keys, last=False):
        '''
        :param keys:
        :type keys: Iterable[str]
        :param last:
        :type last: bool
        :returns: Generator of callables
        :rtype: Generator[
            Callable[[Mapping[Any, Any]], Mapping[Any, Any]],
            Tuple[Iterable[str], bool],
            None
            ]
        '''
        again = True
        handlers = self.handlers
        visited = set()
        while again:
            again = False
            for key in tuple(keys):
                if key in visited:
                    continue

                revisit = False
                for handler in handlers.get(key, ()):
                    if key not in keys:
                        revisit = False
                        break

                    keys, shrunk = (yield lambda data: handler(data, last))
                    revisit |= shrunk

                if revisit:
                    again = True
                else:
                    visited.add(key)

    def _iter_purge(self, keys, last=False):
        '''
        :param keys:
        :type keys: Iterable[str]
        :param last:
        :type last: bool
        :returns: Generator of callables
        :rtype: Generator[
            Callable[[Mapping[Any, Any]], Mapping[Any, Any]],
            Tuple[Iterable[str], bool],
            None
            ]
        '''
        handlers = self.handlers
        for key in tuple(keys):
            if last or key not in handlers:
                yield (lambda data: self._purge(key, data))

    def _iter_tasks(self, keys):
        '''
        Iterate through key handlers, receiving if it worked or not and,
        based on that, keeps returning other handled.

        :param keys: possible handler keys to iterate
        :type keys: Iterable[str]
        :returns: Generator of callables
        :rtype: Generator[
            Callable[[Mapping[Any, Any]], Mapping[Any, Any]],
            Tuple[Iterable[str], bool],
            None
            ]

        Iteration logic:
        1. Handler functions for keys on data.
        2. Functions to purge unhandled keys.
        3. Handler functions for keys on data (receiving last=True)
        4. Functions to purge every key.
        '''
        strategies = [
            lambda keys: self._iter_handled(keys),
            lambda keys: self._iter_purge(keys),
            lambda keys: self._iter_handled(keys, True),
            lambda keys: self._iter_purge(keys, True),
            ]
        for strategy in strategies:
            strategy_iter = strategy(keys)
            try:
                keys, changed = (yield next(strategy_iter))
                while True:
                    keys, changed = (yield strategy_iter.send((keys, changed)))
            except StopIteration:
                pass

    def shrink(self, data, browser, dump_fnc, serialize_fnc):
        '''
        Apply, iteratively, shrink functions to data until it fits into
        given browser limits. If it's not possible, raises error.

        :param data: mapping
        :type data: Mapping
        :param browser: browser limits object
        :type: Browser
        :param dump_fnc: callable which returns binary data payload
        :type dump_fnc: Callable[[Any], int]
        :param serialize_fnc: callable which returns cookies
        :type serialize_fnc: Callable[
            [bytes],
            Iterable[cookieman.cookie.Cookie]
            ]
        :returns: partitioned cookies
        :rtype: Iterable[cookieman.cookie.Cookie]
        :raises exceptions.CookieSizeException: if unable to get under limits
        '''
        handlers_iter = None
        old_size = sys.maxsize
        maxsize = browser.maxsize
        maxcookies = browser.maxcookies

        while True:
            payload = dump_fnc(data)
            cookies = list(serialize_fnc(payload))
            size = sum(map(len, cookies))

            if size <= maxsize and len(cookies) <= maxcookies:
                return cookies

            try:
                if handlers_iter:
                    shrunk = len(payload) < old_size
                    handler = handlers_iter.send((data, shrunk))
                else:
                    handlers_iter = iter(self._iter_tasks(data))
                    handler = next(handlers_iter)
            except StopIteration:
                break

            old_size = len(payload)
            data = handler(data)

        raise cookieman.exceptions.CookieSizeException(
            'Session cookies exceeded browser maxsize %d' % (
                browser.maxsize,
                )
            if len(cookies) > maxsize else
            'Session cookies exceeded browser maxcookies %d' % (
                browser.maxcookies
                )
            )



# -*- coding: utf-8 -*-

import unittest
import random
import string

import six
import pycodestyle
import flask

import cookieman

from . import meta


class TestCodeFormat(six.with_metaclass(meta.TestFileMeta, unittest.TestCase)):
    '''
    pycodestyle unit test
    '''
    meta_module = 'cookieman'
    meta_prefix = 'code'
    meta_file_extensions = ('.py',)

    def meta_test(self, module, filename):
        style = pycodestyle.StyleGuide(quiet=False)
        with self.path(module, filename) as f:
            result = style.check_files([str(f)])
        self.assertFalse(result.total_errors, (
            'Found {errors} code style error{s} (or warning{s}) '
            'on module {module}, file {filename!r}.').format(
                errors=result.total_errors,
                s='s' if result.total_errors > 1 else '',
                module=module,
                filename=filename,
                )
            )


class TestCookieMan(unittest.TestCase):
    module = cookieman

    def setUp(self):
        self.app = flask.Flask('myapp')
        self.app.session_interface = cookieman.CookieMan()
        self.app.secret_key = 'my_app_secret_key'

    def extract_cookies(self, client):
        return [
            c
            for c in client.cookie_jar
            if c.name.startswith(self.app.session_cookie_name)
            ]

    def test_session(self):
        @self.app.route('/write')
        def write():
            flask.session['value'] = 'something'
            return 'OK'

        @self.app.route('/read')
        def read():
            return flask.session.get('value', 'nothing')

        with self.app.test_client() as client:
            response = client.get('/write')
            self.assertEqual(response.data, b'OK')

            response = client.get('/read')
            self.assertEqual(response.data, b'something')
            self.assertIn('Vary', response.headers)
            self.assertNotIn('Set-Cookie', response.headers)

    def test_session_permanent(self):
        @self.app.route('/write')
        def write():
            flask.session['value'] = 'something'
            flask.session.permanent = True
            return 'OK'

        @self.app.route('/read')
        def read():
            return flask.session.get('value', 'nothing')

        @self.app.route('/noop')
        def noop():
            return 'OK'

        self.app.config['SESSION_REFRESH_EACH_REQUEST'] = True

        with self.app.test_client() as client:
            response = client.get('/write')
            self.assertEqual(response.data, b'OK')

            response = client.get('/read')
            self.assertEqual(response.data, b'something')
            self.assertIn('Vary', response.headers)
            self.assertIn('Set-Cookie', response.headers)

            response = client.get('/noop')
            self.assertEqual(response.data, b'OK')
            self.assertNotIn('Vary', response.headers)
            self.assertIn('Set-Cookie', response.headers)

            self.app.config['SESSION_REFRESH_EACH_REQUEST'] = False
            response = client.get('/noop')
            self.assertEqual(response.data, b'OK')
            self.assertNotIn('Vary', response.headers)
            self.assertNotIn('Set-Cookie', response.headers)

    def test_signature(self):
        @self.app.route('/write')
        def write():
            flask.session['value'] = 'something'
            return 'OK'

        @self.app.route('/read')
        def read():
            return flask.session.get('value', 'nothing')

        with self.app.test_client() as client:
            response = client.get('/write')
            self.assertEqual(response.data, b'OK')
            self.app.secret_key = 'other'
            response = client.get('/read')
            self.assertEqual(response.data, b'nothing')

    def test_secret_key(self):
        self.app.secret_key = None
        with self.app.test_request_context():
            with self.assertRaises(RuntimeError):
                flask.session['value'] = 'something'

    def test_removal(self):
        @self.app.route('/set')
        def set():
            flask.session['value'] = 'something'
            return 'OK'

        @self.app.route('/unset')
        def unset():
            del flask.session['value']
            return 'OK'

        with self.app.test_client() as client:
            response = client.get('/set')
            self.assertEqual(response.data, b'OK')
            cookies = self.extract_cookies(client)
            self.assertEqual(len(cookies), 1)

            response = client.get('/unset')
            self.assertEqual(response.data, b'OK')
            cookies = self.extract_cookies(client)
            self.assertEqual(len(cookies), 0)

    def test_shrink(self):
        calls = []

        @self.app.session_interface.register('a')
        def shrink_a(data, last):
            calls.append(len(data['a']))
            if len(data['a']) < 2:
                del data['a']
            else:
                data['a'] = data['a'][:-2]
            return data

        @self.app.route('/')
        def root():
            value = ''.join(
                random.choice(string.ascii_letters)
                for i in range(5000)
                )
            flask.session['a'] = flask.session.get('a', '') + value
            return 'OK'

        with self.app.test_client() as client:
            for i in range(5):
                response = client.get('/')
                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.data, b'OK')

                cookies = self.extract_cookies(client)
                self.assertEqual(len(cookies), 1)
        self.assertGreater(len(calls), 1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import re

from setuptools import setup, find_packages

with io.open('README.rst', 'rt', encoding='utf8') as f:
    readme = f.read()

with io.open('cookieman/__init__.py', 'rt', encoding='utf8') as f:
    version = re.search(r'__version__ = \'(.*?)\'', f.read()).group(1)

setup(
    name='cookieman',
    version=version,
    url='https://gitlab.com/ergoithz/cookieman',
    license='MIT',
    author='Felipe A. Hernandez',
    author_email='ergoithz@gmail.com',
    description='Browser-aware multi-cookie flask session',
    long_description=readme,
    long_description_content_type='text/x-rst',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Framework :: Flask',
        ],
    keywords=['web', 'browser'],
    packages=find_packages(),
    package_data={'cookieman.resources': ['*']},
    install_requires=[
        'flask',
        'werkzeug',
        'msgpack',
        'itsdangerous',
        'importlib_resources',
        ],
    test_suite='cookieman.tests',
    tests_require=['pycodestyle'],
    zip_safe=True,
    platforms='any',
    )



# -*- coding: utf-8 -*-

import functools

import werkzeug.http
import werkzeug.datastructures

import cookieman.exceptions
import cookieman.compat


class Cookie(object):
    header = 'Set-Cookie: '
    mapping_class = werkzeug.datastructures.ImmutableDict
    dump_cookie_fnc = staticmethod(werkzeug.http.dump_cookie)
    byteview_fnc = staticmethod(cookieman.compat.byteview)

    @classmethod
    def default_partition_name_format(cls, name, index):
        return '{}-{:x}'.format(name, index) if index else name

    @property
    def name(self):
        return self._name

    @property
    def value(self):
        return self._value

    @property
    def options(self):
        return self._options

    @property
    def data(self):
        if self._data is None:
            self._data = self.header + self.dump_cookie_fnc(
                self._name, self._value, max_size=0, **self._options)
        return self._data

    def __init__(self, name, value=b'', options={}):
        '''
        :param name: cookie name
        :type name: str
        :param value: cookie value
        :type value: str
        :param options: cookie options
        :type options: collections.abc.Mapping
        '''
        assert isinstance(value, cookieman.compat.BytesLike)
        self._data = None
        self._name = name
        self._value = value
        self._options = self.mapping_class(options)

    def __eq__(self, other):
        return self.data == other.data

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.data)

    def partition(self, maxsize, partition_name_fnc=None):
        '''
        Generate multiple cookies by spliting value in chunks.

        :param maxsize: maximum total cookie size (including header)
        :type maxsize: int
        :param fmt_fnc: cookie name formatting
        :type fmt_fnc: Callable[str, int], str]
        :returns: iterator of cookies
        :rtype: Iterator[cookieman.cookie.Cookie]
        '''
        name_fnc = functools.partial(
            self.default_partition_name_format
            if partition_name_fnc is None else
            partition_name_fnc,
            self.name
            )
        byteview = self.byteview_fnc

        if len(self) < maxsize and name_fnc(0) == self.name:
            yield self
            return

        cls = self.__class__
        total = len(self.value)
        buffer = memoryview(self.value)
        bodysize = maxsize - len(cls('k', b'a', self.options)) + 2

        page = 0
        start = 0
        lastsize = 0

        while start < total or lastsize == maxsize:
            name = name_fnc(page)

            # retry on cookie excess due expansions
            for end in range(start + bodysize - len(name), start, -1):
                cookie = cls(name, byteview(buffer[start:end]), self.options)
                lastsize = len(cookie)

                if lastsize <= maxsize:
                    break
            else:
                raise cookieman.exceptions.CookieSizeException(
                    'Unable to paginate cookies by size %d' % maxsize
                    )

            yield cookie
            start = end
            page += 1


class CookieProcessor(object):
    '''
    Compressed-paginated and signed cookie manager.
    '''
    cookie_class = Cookie

    def __init__(self, session_interface):
        '''
        :param session_interface: flask session interface
        :type session_interface: flask.sessions.SessionInterface
        '''
        self.session_interface = session_interface

    def get_cookie_options(self, app, browser):
        '''
        Get cookie options dict as accepted by flask cookie dump methods.

        :param app: flask app object
        :type app: flask.Flask
        :returns: dict with options
        :rtype: dict
        '''
        return {
            'expires': self.session_interface.get_expiration_time(app),
            'path': self.session_interface.get_cookie_path(app),
            'domain': self.session_interface.get_cookie_domain(app),
            'secure': self.session_interface.get_cookie_secure(app),
            'httponly': self.session_interface.get_cookie_httponly(app),
            }

    def iter_request_cookies(self, app, request, browser):
        '''
        Extract data from cookies.

        :param app: flask app object
        :type app: flask.Flask
        :param request: flask request object
        :type request: flask.Request
        :param browser:
        :type browser: cookieman.limits.Browser
        :returns: iterator of paginated cookie objects
        :rtype: Iterable[cookieman.cookie.Cookie]
        '''
        cookies = request.cookies
        cookie_name = app.session_cookie_name
        cookie_prefix = '%s-' % app.session_cookie_name
        cookies = [
            self.cookie_class(key, value.encode('ascii'))
            for key, value in request.cookies.items()
            if key == cookie_name or key.startswith(cookie_prefix)
            ]

        if cookies:
            cookies.sort(key=lambda x: (len(x.name), x.name))

            maxsize = len(cookies[0])
            for cookie in cookies:
                size = len(cookie)

                # spurious cookie
                if size > maxsize:
                    break

                yield cookie

                # paginated cookies have the same size except for the last one
                if size < maxsize:
                    break

    def iter_response_cookies(self, app, data, browser):
        '''
        Split given byte string for cookies.

        :param app: flask app object
        :type app: flask.Flask
        :param data: byte string
        :type data: bytes
        :param browser:
        :type browser: cookieman.limits.Browser
        :returns: iterator of paginated cookie objects
        :rtype: Iterable[cookieman.cookie.Cookie]
        '''
        options = self.get_cookie_options(app, browser)
        cookie = self.cookie_class(app.session_cookie_name, data, options)
        for page in cookie.partition(browser.maxsize):
            yield page
